import re

from lms.message.definitions.field.lms_field import LMS_FieldMap
from lms.message.tag.lms_tag import (LMS_ControlTag, LMS_DecodedTag,
                                     LMS_EncodedTag, is_tag)
from lms.titleconfig.config import TagConfig


class LMS_MessageText:
    """Class that represents a message text entry."""

    TAG_FORMAT = re.compile(r"(\[[^]]+])")

    def __init__(
        self,
        message: str | list[str | LMS_ControlTag],
        tag_config: TagConfig | None = None,
    ):
        self._tag_config = tag_config

        if isinstance(message, str):
            self._set_parts(message)
        else:
            self._parts = message

    def __iter__(self):
        return iter(self._parts)

    @property
    def text(self) -> str:
        """The raw text of the message."""
        result = []
        for part in self._parts:
            if is_tag(part):
                result.append(part.to_text())
            else:
                result.append(part)
        return "".join(result)

    @text.setter
    def text(self, string: str) -> None:
        self._set_parts(string)

    @property
    def tags(self) -> list[LMS_ControlTag]:
        """The list of control tags in the message."""
        return [part for part in self._parts if is_tag(part)]

    @property
    def tag_positions(self) -> dict[LMS_ControlTag, tuple[int, int]]:
        """Dict of tag objects to their start and end positions in text."""
        positions = {}
        pos = 0
        for part in self._parts:
            text_len = len(part)
            if is_tag(part):
                positions[part] = (pos, pos + text_len)
            pos += text_len
        return positions

    def append_encoded_tag(
        self, group_id: int, tag_index: int, *parameters: str, is_closing: bool = False
    ) -> LMS_EncodedTag:
        """
        Appends an encoded tag to the current message and returns that tag.

        :param group_id: the group index.
        :param tag_index: the index of the tag in the group.
        :param parameters: args of hex strings.
        :param is_closing: whether the tag is closing or not.

        Example
        ------
        >>> message = LMS_MessageText("Text")
        >>> message.append_encoded_tag(0, 3, "00", "23", "43", "32")
        """
        if is_closing:
            tag = LMS_EncodedTag(group_id, tag_index, is_closing=True)
        else:
            tag = LMS_EncodedTag(
                group_id, tag_index, None if not parameters else list(parameters)
            )

        self._parts.append(tag)
        return tag

    def append_decoded_tag(
        self,
        group_name: str,
        tag_name: str,
        is_closing: bool = False,
        **parameters: int | str | float | bool | bytes,
    ) -> LMS_DecodedTag:
        """
        Appends a decoded tag to the current message and returns that tag.

        :param group_name: the group name.
        :param tag_name: the tag name.:
        :param is_closing: whether the tag is closing or not.
        :param parameters: keyword arguments of parameters mapped to their value.

        Example
        -------
        >>> message = LMS_MessageText("Text")
        >>> message.append_decoded_tag("Mii", "Nickname", buffer=1, type="Voice", conversion="None")
        """
        if self._tag_config is None:
            raise ValueError("A TitleConfig is required to append decoded tags.")

        definition = self._tag_config.get_definition_by_names(group_name, tag_name)

        if not parameters:
            tag = LMS_DecodedTag(definition)
        else:
            param_map = LMS_FieldMap.from_dict(parameters, definition.parameters)

            if is_closing:
                tag = LMS_DecodedTag(definition, is_closing=True)
            else:
                tag = LMS_DecodedTag(definition, param_map)

        self._parts.append(tag)
        return tag

    def append_tag_string(self, tag: str) -> LMS_ControlTag:
        """
        Appends a tag to the current message given a string.

        :param tag: the tag string.
        """
        if re.fullmatch(LMS_DecodedTag.TAG_FORMAT, tag):
            if self._tag_config is None:
                raise ValueError("TagConfig is required to append decoded tags.")
            tag_obj = LMS_DecodedTag.from_string(tag, self._tag_config)
        elif re.fullmatch(LMS_EncodedTag.TAG_FORMAT, tag):
            tag_obj = LMS_EncodedTag.from_string(tag)
        else:
            raise ValueError(f"Invalid format for tag '{tag}'.")

        self._parts.append(tag_obj)
        return tag_obj

    def _set_parts(self, text: str) -> None:
        self._parts = []
        for part in self.TAG_FORMAT.split(text):
            if bool(re.match(self.TAG_FORMAT, part)):
                self.append_tag_string(part)
            else:
                self._parts.append(part)



import os
from typing import BinaryIO, cast

from lms.common.lms_datatype import LMS_DataType
from lms.common.stream.fileinfo import read_file_info
from lms.common.stream.hashtable import read_labels
from lms.common.stream.section import read_section_data
from lms.fileio.io import FileReader
from lms.project.msbp import MSBP
from lms.project.section.ali2 import read_ali2
from lms.project.section.ati2 import read_ati2
from lms.project.section.clr1 import read_clr1
from lms.project.section.string import read_strings
from lms.project.section.syl3 import read_styles
from lms.project.section.tag2 import read_tag2
from lms.project.section.tgg2 import read_tgg2
from lms.project.section.tgp2 import read_tgp2
from PylibMS.lms.project.definitions.color import LMS_Color

__all__ = ["read_msbp", "read_msbp_path"]


def read_msbp_path(file_path: str) -> MSBP:
    """
    Reads and retrieves a MSBP file from a given path.

    :param file_path: the path to the MSBP file.

    Example
    ---------
    >>> msbp = read_msbp_path("path/to/file.msbp")
    """
    with open(file_path, "rb") as stream:
        return read_msbp(stream)


def read_msbp(stream: BinaryIO | bytes) -> MSBP:
    """
    Reads and retrieves a MSBP file from a specified stream.

    :param stream: an ``IOBase``, ``BytesIO``, ``memoryview``, or ``bytes`` object.

    Example
    ---------
    >>> msbp = read_msbp(stream)
    """
    reader = FileReader(stream)
    file_info = read_file_info(reader, MSBP.MAGIC)

    colors = None
    attr_definitions = None
    attribute_list_items = None
    tag_groups = None
    tag_definitions = None
    tag_param_definitions = None
    tag_list_items = None
    styles = None
    source_list = None

    items = []
    for magic, _ in read_section_data(reader, file_info.section_count):
        match magic:
            case "CLB1" | "ALB1" | "SLB1":
                # Set the name attributes of last read item
                labels, _ = read_labels(reader)
                for i in labels:
                    items[i].name = labels[i]
            case "CLR1":
                colors = read_clr1(reader)
                items = colors
            case "ATI2":
                attr_definitions = read_ati2(reader)
                items = attr_definitions
            case "ALI2":
                attribute_list_items = read_ali2(reader)
            case "TGG2":
                tag_groups = read_tgg2(reader, file_info.version)
            case "TAG2":
                tag_definitions = read_tag2(reader)
            case "TGP2":
                tag_param_definitions = read_tgp2(reader)
            case "TGL2":
                tag_list_items = read_strings(reader, False)
            case "SYL3":
                styles = read_styles(reader)
                items = styles
            case "CTI1":
                source_list = read_strings(reader, True)
            case _:
                raise ValueError(f"Unknown section magic '{magic}' in MSBP file.")

    if attr_definitions is not None:
        for definition in attr_definitions:
            if definition.datatype is LMS_DataType.LIST:
                definition.list_items = attribute_list_items[definition.list_index]

    if tag_groups is not None:
        for group in tag_groups:
            group.set_all_definitions(
                tag_definitions, tag_param_definitions, tag_list_items
            )

    file = MSBP(file_info, colors, attr_definitions, tag_groups, styles, source_list)

    if isinstance(stream, BinaryIO):
        file.name = os.path.basename(stream.name).removesuffix(".msbp")

    return file



from __future__ import annotations

from dataclasses import dataclass, field

from lms.common.lms_datatype import LMS_DataType
from lms.titleconfig.definitions.value import ValueDefinition


class TagConfig:
    """Class that represents a tag structure definition."""

    def __init__(self, group_map: dict[int, str], definitions: list[TagDefinition]):
        self._group_map = group_map
        self._definitions = definitions

    @property
    def group_map(self) -> dict[int, str]:
        return self._group_map

    @property
    def definitions(self) -> list[TagDefinition]:
        return self._definitions

    def get_definition_by_names(self, group_name: str, tag_name: str) -> TagDefinition:
        group_index = None
        for i, name in self.group_map.items():
            if name == group_name:
                group_index = i
                break

        if group_index is None:
            raise KeyError(
                f"Group name '{group_name}' was not found! Is the group defined?"
            )

        for tag_def in self.definitions:
            if tag_def.group_id == group_index and tag_def.tag_name == tag_name:
                return tag_def

        raise KeyError(
            f"Tag name '{tag_name}' not found in group '{group_name}'. Is the tag defined?"
        )

    def get_definition_by_indexes(
        self, group_id: int, tag_index: int
    ) -> TagDefinition | None:
        if group_id not in self._group_map:
            return None

        for tag_def in self.definitions:
            if tag_def.group_id == group_id and tag_def.tag_index == tag_index:
                return tag_def

        return None


@dataclass(frozen=True)
class TagDefinition:
    """Class that represents a signle definition in the tag config."""

    group_name: str
    group_id: int
    tag_name: str
    tag_index: int
    description: str
    parameters: list[ValueDefinition] = field(default_factory=list)

    @classmethod
    def from_dict(cls, data: dict, group_map: dict[int, str]):
        tag_name = data["name"]
        group_id, tag_index = data["group_id"], data["tag_index"]
        description = data["description"]
        group_name = group_map[group_id]

        if "parameters" not in data:
            return cls(group_name, group_id, tag_name, tag_index, description)

        parameters = []
        for param_def in data["parameters"]:
            name = param_def["name"]
            param_description = param_def["description"]
            datatype = LMS_DataType.from_string(param_def["datatype"])
            list_items = param_def.get("list_items")
            parameters.append(
                ValueDefinition(name, param_description, datatype, list_items)
            )

        return cls(
            group_name,
            group_id,
            tag_name,
            tag_index,
            description,
            parameters,
        )



from dataclasses import dataclass

from lms.fileio.encoding import FileEncoding


@dataclass
class LMS_FileInfo:
    is_big_endian: bool = False
    encoding: FileEncoding = FileEncoding.UTF16
    version: int = 3
    section_count: int = 2



from lms.fileio.io import FileReader


def read_strings(reader: FileReader, four_byte_count: bool) -> list[str]:
    string_list = []

    if four_byte_count:
        count = reader.read_uint32()
    else:
        count = reader.read_uint16()
        reader.skip(2)

    for offset in reader.read_offset_array(count):
        reader.seek(offset)
        string_list.append(reader.read_encoded_string())

    return string_list



from lms.message.definitions.field.lms_field import LMS_FieldMap
from lms.message.definitions.lms_messagetext import LMS_MessageText
from lms.titleconfig.definitions.attribute import AttributeConfig
from lms.titleconfig.definitions.tags import TagConfig


class MSBTEntry:
    """A class that represents an entry in a MSBT file."""

    def __init__(
        self,
        name: str,
        *,
        message: LMS_MessageText | str | None = "",
        attribute: LMS_FieldMap | bytes | None = None,
        style_index: int | None = None,
    ):
        self.name = name

        if not isinstance(message, (LMS_MessageText, str)):
            raise TypeError(
                f"An invalid type was provided for text in entry '{name}'! Expected LMS_MessageText object or str got {type(message)}"
            )

        if isinstance(message, str):
            self._message = LMS_MessageText(message)
        else:
            self._message = message

        if attribute is not None and not isinstance(attribute, (LMS_FieldMap, bytes)):
            raise TypeError(
                f"An invalid type was provided for attribute in entry '{name}'. "
                f"Expected LMS_FieldMap or bytes, got {type(attribute)}"
            )

        self._attribute = attribute
        self.style_index = style_index

    @property
    def message(self) -> LMS_MessageText | None:
        """The message object for the instance."""
        return self._message

    @property
    def attribute(self) -> LMS_FieldMap | bytes | None:
        """The attribute for the instance."""
        return self._attribute

    def to_dict(self) -> dict:
        """Converts the MSBTEntry instance into a dictionary object."""
        result: dict[str, int | str | dict | None] = {
            "name": self.name,
            "message": "" if self.message is None else self.message.text,
        }

        if self._attribute is not None:
            if isinstance(self._attribute, bytes):
                result["attribute"] = self._attribute.hex().upper()
            else:
                result["attribute"] = self._attribute.to_dict()

        result["style_index"] = self.style_index
        return result

    @classmethod
    def from_dict(
        cls,
        data: dict,
        attribute_config: AttributeConfig | None = None,
        tag_config: TagConfig | None = None,
    ):
        """
        Creates a MSBTEntry from a dictionary object.

        :param data: the dictionary data.
        :param attribute_config: the config to use to import decoded attributes.
        :param tag_config: the config to use if decoded tags are included in the message.
        """
        message = data.get("message", "")
        attribute = data.get("attribute")
        style_index = data.get("style_index")

        if tag_config is not None:
            message = LMS_MessageText(message, tag_config)

        if attribute is not None:
            if not isinstance(attribute, (dict, str)):
                raise TypeError("Invalid attribute type provided!")

            if isinstance(attribute, dict):
                if attribute_config is None:
                    raise TypeError(
                        "A valid attribute config must be provided for decoded attributes!"
                    )
                attribute = LMS_FieldMap.from_dict(
                    attribute, attribute_config.definitions
                )
            else:
                attribute = bytes.fromhex(attribute)

        return cls(
            data["name"], message=message, attribute=attribute, style_index=style_index
        )



from lms.common.lms_fileinfo import LMS_FileInfo
from lms.project.definitions.attribute import LMS_AttributeDefinition
from lms.project.definitions.color import LMS_Color
from lms.project.definitions.style import LMS_Style
from lms.project.definitions.tag import LMS_TagGroup


class MSBP:
    """
    A class that represents a MSBP file.

    https://nintendo-formats.com/libs/lms/msbp.html.
    """

    MAGIC = "MsjPrjBn"

    def __init__(
        self,
        info: LMS_FileInfo,
        colors: list[LMS_Color] | None,
        config: list[LMS_AttributeDefinition] | None,
        tag_groups: list[LMS_TagGroup] | None,
        styles: list[LMS_Style] | None,
        source_files: list[str] | None,
    ):
        self.name: str = ""

        self._info = info

        self._colors = colors
        self._attribute_definitions = config
        self._tag_groups = tag_groups
        self._styles = styles
        self._source_files = source_files

    @property
    def info(self) -> LMS_FileInfo:
        """The stream info of the MSBP."""
        return self._info

    @property
    def colors(self) -> list[LMS_Color] | None:
        """The color definitions for the project."""
        return self._colors

    @property
    def attribute_definitions(self) -> list[LMS_AttributeDefinition] | None:
        """The attribute definitions for the project instance."""
        return self._attribute_definitions

    @property
    def tag_groups(self) -> list[LMS_TagGroup] | None:
        """The tag group definitions for the project instance."""
        return self._tag_groups

    @property
    def style_list(self) -> list[LMS_Style] | None:
        """The style definitions for the project instance."""
        return self._styles

    @property
    def source_files(self) -> list[str] | None:
        """The source file definitions for the project instance."""
        return self._source_files



from lms.common.lms_datatype import LMS_DataType
from lms.fileio.io import FileReader
from lms.project.definitions.tag import LMS_TagParamDefinition


def read_tgp2(reader: FileReader) -> list[LMS_TagParamDefinition]:
    parameter_info = []

    count = reader.read_uint32()
    for offset in reader.read_offset_array(count):
        reader.seek(offset)

        datatype = LMS_DataType(reader.read_uint8())

        if datatype is not LMS_DataType.LIST:
            name = reader.read_encoded_string()
            parameter_info.append(LMS_TagParamDefinition(name, datatype))
            continue

        reader.skip(1)
        list_count = reader.read_uint16()
        list_indexes = reader.read_uint16_array(list_count)
        name = reader.read_encoded_string()
        parameter_info.append(
            LMS_TagParamDefinition(name, LMS_DataType.LIST, list_indexes)
        )

    return parameter_info



from __future__ import annotations

from dataclasses import dataclass, field

from lms.common.lms_datatype import LMS_DataType


class LMS_TagGroup:
    def __init__(
        self,
        name: str,
        group_id: int,
        tag_indexes: list[int],
    ):
        self._name = name
        self._id = group_id
        self._tag_indexes = tag_indexes

        self.tag_definitions: list[LMS_TagDefinition] = []

    @property
    def name(self) -> str:
        return self._name

    @property
    def group_id(self) -> int:
        return self._id

    def set_all_definitions(
        self,
        tag_definitions: list["LMS_TagDefinition"],
        parameter_definitions: list["LMS_TagParamDefinition"],
        list_items: list[list[str]],
    ) -> None:

        self.tag_definitions.extend(tag_definitions[i] for i in self._tag_indexes)
        for tag in self.tag_definitions:
            tag.parameter_definitions.extend(
                parameter_definitions[i] for i in tag.parameter_indexes
            )
            for parameter in tag.parameter_definitions:
                if parameter.datatype is LMS_DataType.LIST:
                    parameter.list_items = [
                        item for i in parameter.list_indexes for item in list_items[i]
                    ]


class LMS_TagDefinition:
    def __init__(
        self,
        name: str,
        parameter_indexes: list[int],
        parameter_definitions: list[LMS_TagParamDefinition] | None = None,
    ):
        self._name = name
        self._parameter_indexes = (
            parameter_indexes if parameter_indexes is not None else []
        )
        self.parameter_definitions = (
            parameter_definitions if parameter_definitions is not None else []
        )

    @property
    def name(self) -> str:
        return self._name

    @property
    def parameter_indexes(self) -> list[int]:
        return self._parameter_indexes


class LMS_TagParamDefinition:
    def __init__(
        self,
        name: str,
        datatype: LMS_DataType,
        list_indexes: list[int] | None = None,
    ):
        self._name = name
        self.list_items: list[str] = []

        self._datatype = datatype
        self._list_indexes = list_indexes if list_indexes is not None else []

    @property
    def name(self) -> str:
        return self._name

    @property
    def datatype(self) -> LMS_DataType:
        return self._datatype

    @property
    def list_indexes(self) -> list[int]:
        return self._list_indexes



from importlib import resources

import yaml

from lms.common.lms_datatype import LMS_DataType
from lms.project.msbp import MSBP
from lms.titleconfig.definitions.attribute import AttributeConfig
from lms.titleconfig.definitions.tags import TagConfig, TagDefinition
from lms.titleconfig.definitions.value import ValueDefinition


class TitleConfig:
    """Represents a configuration for a specific title."""

    TAG_KEY = "tag_definitions"
    ATTR_KEY = "attribute_definitions"

    PRESET_LIST = [
        file.name.removesuffix(".yaml")
        for file in resources.files("lms.titleconfig.presets").iterdir()
    ]

    def __init__(
        self,
        game: str | None,
        attribute_config_map: dict[str, AttributeConfig] | None = None,
        tag_config: TagConfig | None = None,
    ):
        self._game = game
        self._attribute_config_map = attribute_config_map
        self._tag_config = tag_config

    @property
    def game(self) -> str | None:
        """The name of the game for the titleconfig."""
        return self._game

    @property
    def tag_config(self) -> TagConfig | None:
        """The loaded tag config instance."""
        return self._tag_config

    @property
    def attribute_configs(self) -> tuple[AttributeConfig, ...]:
        """Returns a tuple of all attribute configurations in read-only form."""
        if self._attribute_config_map is None:
            return ()
        return tuple(self._attribute_config_map.values())

    def get_attribute_config(self, name: str) -> AttributeConfig:
        """
        Returns the attribute configuration given the name.

        :param name: the name of the attribute config.
        """
        if self._attribute_config_map is None:
            raise ValueError(f"There are no attribute configs!")

        if name not in self._attribute_config_map:
            raise KeyError(
                f"The attribute config '{name}' does not exist in the TitleConfig!"
            )

        return self._attribute_config_map[name]

    @classmethod
    def load_preset(cls, game: str):
        """
        Loads an existing preset from a game.

        :param game: the game preset.

        The list of presets are found with `TitleConfig.preset_list`.
        """
        preset_map = {preset.lower(): preset for preset in cls.PRESET_LIST}

        if game.lower() not in preset_map:
            raise FileNotFoundError(f"Preset '{game}' not found.")

        actual_name = preset_map[game.lower()]

        with resources.open_text("lms.titleconfig.presets", f"{actual_name}.yaml") as f:
            return cls.load_config(f.read())

    @classmethod
    def load_file(cls, file_path: str):
        """
        Loads a config from a file.

        :param file_path: the path to the config.
        """
        with open(file_path, "r") as f:
            return TitleConfig.load_config(f.read())

    @classmethod
    def load_config(cls, content: str | dict):
        """
        Loads the config of a specified game.

        :param content: the config content, as a string or loaded as a dictionary.
        """

        if isinstance(content, str):
            parsed_content = yaml.safe_load(content)
        else:
            parsed_content = content

        game = parsed_content.get("game")

        attribute_configs = {}
        for config in parsed_content[cls.ATTR_KEY]:
            definitions = [
                ValueDefinition.from_dict(value_def)
                for value_def in config["definitions"]
            ]
            attribute_configs[config["name"]] = AttributeConfig(
                config["name"], config["description"], definitions
            )

        tag_definitions = []
        group_map = parsed_content[cls.TAG_KEY]["groups"]

        for tag_def in parsed_content[cls.TAG_KEY]["tags"]:
            tag_definitions.append(TagDefinition.from_dict(tag_def, group_map))

        tag_config = TagConfig(group_map, tag_definitions)

        return cls(game, attribute_configs, tag_config)

    @staticmethod
    def generate_file(file_path: str, game: str, project: MSBP) -> None:
        """
        Generates a title config file for a specific game.

        :param file_path: the path to the yaml file.
        :param game: the name of the game to create the config for.
        :param project: a MSBP object.
        """
        with open(file_path, "w+") as f:
            yaml.safe_dump(
                TitleConfig.generate_config(game, project),
                f,
                default_flow_style=False,
                sort_keys=False,
            )

    @staticmethod
    def generate_config(game: str, project: MSBP) -> dict | None:
        """
        Generates a title config file for the specified game.

        :param game: the name of the game to create the config for.
        :param project: a MSBP object.
        """
        config = {}
        config["game"] = game

        if project.tag_groups is not None:
            config[TitleConfig.TAG_KEY] = {
                "groups": {group.group_id: group.name for group in project.tag_groups},
                "tags": [],
            }

            for group in project.tag_groups:
                for i, tag_def in enumerate(group.tag_definitions):

                    definition = {
                        "name": tag_def.name,
                        "group_id": group.group_id,
                        "tag_index": i,
                        "description": "",
                        "parameters": None if not tag_def.parameter_definitions else [],
                    }

                    if definition["parameters"] is None:
                        continue

                    for param_def in tag_def.parameter_definitions:
                        param_definition: dict[str, str | list] = {
                            "name": param_def.name,
                            "description": "",
                            "datatype": param_def.datatype.to_string(),
                        }

                        if param_def.datatype is LMS_DataType.LIST:
                            param_definition["list_items"] = param_def.list_items

                        definition["parameters"].append(param_definition)

                    config[TitleConfig.TAG_KEY]["tags"].append(definition)

        config[TitleConfig.ATTR_KEY] = []
        if project.attribute_definitions is not None:

            attr_definitions = []
            for attr_def in project.attribute_definitions:
                definition = {
                    "name": attr_def.name,
                    "description": "",
                    "datatype": attr_def.datatype.to_string(),
                }
                if attr_def.datatype is LMS_DataType.LIST:
                    definition["list_items"] = attr_def.list_items

                attr_definitions.append(definition)

            # Main attribute entries
            config[TitleConfig.ATTR_KEY].append(
                {
                    "name": project.name,
                    "description": "",
                    "definitions": attr_definitions,
                }
            )

        return config



from dataclasses import dataclass

from lms.common.lms_datatype import LMS_DataType
from lms.fileio.io import FileReader, FileWriter
from lms.message.definitions.field.io import read_field, write_field
from lms.message.definitions.field.lms_field import (LMS_DataType, LMS_Field,
                                                     LMS_FieldMap)
from lms.titleconfig.definitions.attribute import AttributeConfig


@dataclass(frozen=True)
class ATR1Data:
    attributes: list[bytes] | list[LMS_FieldMap]
    size_per_attribute: int
    string_table: bytes | None


def read_atr1(
    reader: FileReader, config: AttributeConfig | None, section_size: int
) -> ATR1Data:
    if config is None:
        return read_encoded_atr1(reader, section_size)
    return read_decoded_atr1(reader, config)


def read_encoded_atr1(reader: FileReader, section_size: int) -> ATR1Data:
    absolute_size = section_size + reader.tell()

    attribute_count = reader.read_uint32()
    size_per_attribute = reader.read_uint32()

    attributes = [reader.read_bytes(size_per_attribute) for _ in range(attribute_count)]

    string_table = None
    if section_size > 8 + size_per_attribute * attribute_count:
        string_table = reader.read_bytes(absolute_size - reader.tell())

    return ATR1Data(attributes, size_per_attribute, string_table)


def read_decoded_atr1(reader: FileReader, config: AttributeConfig) -> ATR1Data:
    section_start = reader.tell()

    attr_count = reader.read_uint32()
    size_per_attribute = reader.read_uint32()

    attributes, attr_start = [], reader.tell()

    for i in range(attr_count):
        reader.seek(attr_start + i * size_per_attribute)

        attribute = {}
        for definition in config.definitions:
            if definition.datatype is LMS_DataType.STRING:
                last = reader.tell() + 4
                reader.seek(section_start + reader.read_uint32())
                value = LMS_Field(reader.read_encoded_string(), definition)
                reader.seek(last)
            else:
                value = read_field(reader, definition)

            attribute[definition.name] = value

        attributes.append(LMS_FieldMap(attribute))

    return ATR1Data(attributes, size_per_attribute, None)


def write_encoded_atr1(
    writer: FileWriter,
    attributes: list[bytes],
    size_per_attribute: int,
    string_table: bytes | None,
):
    writer.write_uint32(len(attributes))

    if attributes:
        writer.write_uint32(size_per_attribute)
    else:
        writer.write_uint32(0)

    for attr in attributes:
        writer.write_bytes(attr)

    if string_table is not None:
        writer.write_bytes(string_table)


def write_decoded_atr1(
    writer: FileWriter, attributes: list[LMS_FieldMap], size_per_attribute: int
) -> None:
    writer.write_uint32(len(attributes))
    writer.write_uint32(size_per_attribute)

    string_table = []
    string_offset = 8 + size_per_attribute * len(attributes)

    for attr in attributes:
        for field in attr:
            if field.datatype is LMS_DataType.STRING:
                field.value = str(field.value)

                string_table.append(field.value)
                writer.write_uint32(string_offset)
                string_offset += len(field.value) * writer.encoding.width + len(
                    writer.encoding.terminator
                )
            else:
                write_field(writer, field)

    for string in string_table:
        writer.write_encoded_string(string)



from typing import cast

from lms.common.lms_datatype import LMS_DataType, is_string_datatype
from lms.fileio.io import FileReader, FileWriter
from lms.message.definitions.field.io import read_field, write_field
from lms.message.definitions.field.lms_field import LMS_Field, LMS_FieldMap
from lms.message.tag.lms_tagexceptions import (LMS_TagReadingError,
                                               LMS_TagWritingException)
from lms.titleconfig.definitions.tags import TagDefinition

TAG_PADDING_BYTE = b"\xcd"


def read_encoded_parameters(
    reader: FileReader, parameter_size: int
) -> list[str] | None:
    hex_parameters = reader.read_bytes(parameter_size).hex().upper()
    encoded_parameters = [
        f"{hex_parameters[i : i + 2]}" for i in range(0, len(hex_parameters), 2)
    ]
    return encoded_parameters


def read_decoded_parameters(
    reader: FileReader, definition: TagDefinition
) -> LMS_FieldMap | None:
    parameters = {}
    for param in definition.parameters:
        param_offset = reader.tell()
        try:
            if param.datatype is LMS_DataType.STRING:
                value = LMS_Field(reader.read_len_string_encoded(), param)
            else:
                value = read_field(reader, param)
        except Exception as e:
            raise LMS_TagReadingError(
                f"An error occurred reading tag '[{definition.group_name}:{definition.tag_name}]', parameter '{param.name}' at offset {param_offset}"
            ) from e

        parameters[param.name] = value

    return LMS_FieldMap(parameters)


def write_encoded_parameters(writer: FileWriter, parameters: list[str]) -> None:
    writer.write_uint16(len(parameters))
    for param in parameters:
        writer.write_bytes(bytes.fromhex(param))


def write_decoded_parameters(
    writer: FileWriter, parameters: LMS_FieldMap, group_name: str, tag_name: str
) -> None:
    param_size = 0

    for field in parameters:
        if is_string_datatype(field.value, field.datatype):
            param_size += 2 + len(field.value) * writer.encoding.width
        else:
            param_size += field.datatype.stream_size

    if needs_padding := (param_size % writer.encoding.width != 0):
        param_size += 1

    writer.write_uint16(param_size)

    for field in parameters:
        try:
            if is_string_datatype(field.value, field.datatype):
                # Tags are padded by a 0xCD byte if the size is not aligned to the encoding
                # This can occur before a string parameter, or at the end of the tag.
                # If a string parameter exists, then the padding will always be at the first string instance
                if needs_padding:
                    writer.write_bytes(TAG_PADDING_BYTE)
                    needs_padding = False

                writer.write_len_encoded_string(field.value)
            else:
                write_field(writer, field)
        except Exception as e:
            raise LMS_TagWritingException(
                f"An error occurred in tag [{group_name}:{tag_name} writing parameter '{field.name}' at offset {writer.tell()}!"
            ) from e

    if needs_padding:
        writer.write_bytes(TAG_PADDING_BYTE)



from lms.fileio.encoding import FileEncoding
from lms.fileio.io import FileReader, FileWriter
from lms.message.tag.io.param_io import (read_decoded_parameters,
                                         read_encoded_parameters,
                                         write_decoded_parameters,
                                         write_encoded_parameters)
from lms.message.tag.lms_tag import (LMS_ControlTag, LMS_DecodedTag,
                                     LMS_EncodedTag)
from lms.message.tag.lms_tagexceptions import LMS_TagReadingError
from lms.titleconfig.definitions.tags import TagConfig, TagDefinition

TAG_PADDING_BYTE = b"\xcd"


def get_tag_indicator(encoding: FileEncoding, is_big_endian: bool):
    start_indicator = int.to_bytes(
        0x0E, encoding.width, "little" if not is_big_endian else "big"
    )
    closing_indicator = int.to_bytes(
        0x0F, encoding.width, "little" if not is_big_endian else "big"
    )
    return start_indicator, closing_indicator


def read_tag(
    reader: FileReader,
    tag_config: TagConfig | None,
    is_closing: bool,
    suppress_tag_errors: bool,
) -> LMS_ControlTag:
    group_id = reader.read_uint16()
    tag_index = reader.read_uint16()
    start = reader.tell()

    if tag_config is None:
        return _read_encoded_tag(reader, group_id, tag_index, is_closing)

    definition = tag_config.get_definition_by_indexes(group_id, tag_index)

    # Tags not defined in the config are not considered fallback tags
    # Not all configs will define every tag, so this is simply a measure to still read tags that aren't defined
    if definition is None:
        return _read_encoded_tag(reader, group_id, tag_index, is_closing)

    # There doesn't need to be error handling in this case since there are no parameters for closing tags
    if is_closing:
        return _read_decoded_tag(reader, definition, is_closing=True)

    try:
        tag = _read_decoded_tag(reader, definition)
    except LMS_TagReadingError as e:
        if not suppress_tag_errors:
            raise e

        reader.seek(start)
        return _read_encoded_tag(reader, group_id, tag_index, is_fallback=True)

    return tag


def _read_encoded_tag(
    reader: FileReader,
    group_id: int,
    tag_index: int,
    is_closing: bool = False,
    is_fallback: bool = False,
) -> LMS_EncodedTag:
    if is_closing:
        return LMS_EncodedTag(group_id, tag_index, is_closing=True)

    if not (parameter_size := reader.read_uint16()):
        return LMS_EncodedTag(group_id, tag_index)

    parameters = read_encoded_parameters(reader, parameter_size)
    return LMS_EncodedTag(group_id, tag_index, parameters, is_fallback)


def _read_decoded_tag(
    reader: FileReader, definition: TagDefinition, is_closing: bool = False
) -> LMS_DecodedTag:
    parameter_size = reader.read_uint16()
    end = reader.tell() + parameter_size

    if not parameter_size:
        return LMS_DecodedTag(definition)

    if is_closing:
        return LMS_DecodedTag(definition, is_closing=True)

    parameters = read_decoded_parameters(reader, definition)
    reader.seek(end)
    return LMS_DecodedTag(definition, parameters)


def write_tag(writer: FileWriter, tag: LMS_EncodedTag | LMS_DecodedTag) -> None:
    start_indicator, close_indicator = get_tag_indicator(
        writer.encoding, writer.is_big_endian
    )
    writer.write_bytes(start_indicator if not tag.is_closing else close_indicator)

    writer.write_uint16(tag.group_id)
    writer.write_uint16(tag.tag_index)

    if tag.is_closing:
        return

    if tag.parameters is None:
        writer.write_uint16(0)
        return

    if isinstance(tag, LMS_EncodedTag):
        write_encoded_parameters(writer, tag.parameters)
    else:
        write_decoded_parameters(writer, tag.parameters, tag.group_name, tag.tag_name)



from lms.common.lms_datatype import LMS_DataType


class LMS_AttributeDefinition:
    def __init__(self, datatype: LMS_DataType, offset: int, list_index: int):
        self.name: str | None = None

        self.list_items: list[str] = []

        self._datatype = datatype
        self._offset = offset
        self._list_index = list_index

    @property
    def datatype(self) -> LMS_DataType:
        return self._datatype

    @property
    def offset(self) -> int:
        return self._offset

    @property
    def list_index(self) -> int:
        return self._list_index



from typing import Callable

from lms.common.lms_datatype import (LMS_DataType, is_bool_datatype,
                                     is_bytes_datatype, is_list_datatype,
                                     is_number_datatype)
from lms.fileio.io import FileReader, FileWriter
from lms.message.definitions.field.lms_field import LMS_Field
from lms.titleconfig.definitions.value import ValueDefinition


def read_field(reader: FileReader, definition: ValueDefinition) -> LMS_Field:
    # String is excluded as their reading varies between tags and attributes
    match definition.datatype:
        case LMS_DataType.UINT8:
            value = reader.read_uint8()
        case LMS_DataType.UINT16:
            value = reader.read_uint16()
        case LMS_DataType.UINT32:
            value = reader.read_uint32()
        case LMS_DataType.INT8:
            value = reader.read_int8()
        case LMS_DataType.INT16:
            value = reader.read_int16()
        case LMS_DataType.INT32:
            value = reader.read_int32()
        case LMS_DataType.FLOAT32:
            value = reader.read_float32()
        case LMS_DataType.LIST:
            index = reader.read_uint8()
            value = definition.list_items[index]
        case LMS_DataType.BOOL:
            value = bool(reader.read_uint8())
        case LMS_DataType.BYTES:
            value = reader.read_bytes(1)

    return LMS_Field(value, definition)  # type: ignore


def write_field(writer: FileWriter, field: LMS_Field) -> None:
    if is_number_datatype(field.value, field.datatype):
        write_functions: dict[LMS_DataType, Callable] = {
            LMS_DataType.UINT8: writer.write_uint8,
            LMS_DataType.INT8: writer.write_int8,
            LMS_DataType.UINT16: writer.write_uint16,
            LMS_DataType.INT16: writer.write_int16,
            LMS_DataType.UINT32: writer.write_uint32,
            LMS_DataType.INT32: writer.write_int32,
            LMS_DataType.FLOAT32: writer.write_float32,
        }
        write_functions[field.datatype](field.value)
        return

    if is_list_datatype(field.value, field.datatype):
        writer.write_uint8(field.list_items.index(field.value))
    elif is_bytes_datatype(field.value, field.datatype):
        writer.write_bytes(field.value)
    elif is_bool_datatype(field.value, field.datatype):
        writer.write_uint8(bool(field.value))
    else:
        raise ValueError(f"Unsupported datatype: {field.datatype}")



from lms.common import lms_exceptions
from lms.common.lms_fileinfo import LMS_FileInfo
from lms.fileio.encoding import FileEncoding
from lms.fileio.io import FileReader, FileWriter

DATA_START = 0x20
LITTLE_ENDIAN_BOM = b"\xff\xfe"
BIG_ENDIAN_BOM = b"\xfe\xff"


def read_file_info(reader: FileReader, expected_magic: str) -> LMS_FileInfo:
    magic = reader.read_string_len(8)

    if magic != expected_magic:
        raise lms_exceptions.LMS_UnexpectedMagicError(
            f"Invalid magic!' Expected {expected_magic}', got '{magic}'."
        )

    is_big_endian = reader.read_bytes(2) == BIG_ENDIAN_BOM
    reader.is_big_endian = is_big_endian

    reader.skip(2)

    encoding = FileEncoding(reader.read_uint8())
    reader.encoding = encoding

    version = reader.read_uint8()
    section_count = reader.read_uint16()

    reader.skip(2)
    file_size = reader.read_uint32()

    reader.seek(0, 2)
    if file_size != reader.tell():
        raise lms_exceptions.LMS_MisalignedSizeError(f"Filesize is misaligned!")

    reader.seek(DATA_START)

    return LMS_FileInfo(
        is_big_endian,
        encoding,
        version,
        section_count,
    )


def write_file_info(writer: FileWriter, magic: str, file_info: LMS_FileInfo) -> None:
    writer.is_big_endian = file_info.is_big_endian
    writer.encoding = file_info.encoding

    writer.write_string(magic)
    writer.write_bytes(
        LITTLE_ENDIAN_BOM if not file_info.is_big_endian else BIG_ENDIAN_BOM
    )
    writer.write_bytes(b"\x00\x00")

    writer.write_uint8(file_info.encoding.value)
    writer.write_uint8(file_info.version)
    writer.write_uint16(file_info.section_count)

    writer.write_bytes(b"\x00\x00")
    writer.write_bytes(b"\x00" * 4)
    writer.write_bytes(b"\x00" * 10)
    writer.seek(DATA_START)



from lms.fileio.io import FileReader, FileWriter


def read_labels(reader: FileReader) -> tuple[dict, int]:
    labels = {}

    data_start = reader.tell()
    slot_count = reader.read_uint32()
    for _ in range(slot_count):
        label_count = reader.read_uint32()
        offset = reader.read_uint32()
        next_offset = reader.tell()

        reader.seek(data_start + offset)
        for _ in range(label_count):
            length = reader.read_uint8()
            label = reader.read_string_len(length)
            item_index = reader.read_uint32()
            labels[item_index] = label

        reader.seek(next_offset)

    sorted_labels = {i: labels[i] for i in sorted(labels)}

    # While the slot count is consistent for most files, some vary them.
    # Return the slot count to ensure that this change can be recorded.
    return sorted_labels, slot_count


def write_labels(writer: FileWriter, labels: list[str], slot_count: int) -> None:
    writer.write_uint32(slot_count)

    hash_slots = {}
    index_map = {}
    # Add each label to each hash slot
    for i, label in enumerate(labels):
        hash = _calculate_hash(label, slot_count)

        if hash not in hash_slots:
            hash_slots[hash] = [label]
        else:
            hash_slots[hash].append(label)

        index_map[label] = i

    label_offsets = slot_count * 8 + 4

    hash_slots = dict(sorted(hash_slots.items(), key=lambda x: x[0]))

    for i in range(slot_count):
        if i in hash_slots:
            writer.write_uint32(len(hash_slots[i]))
            writer.write_uint32(label_offsets)
            for label in hash_slots[i]:
                label_offsets += len(label) + 5
        else:
            writer.write_uint32(0)
            writer.write_uint32(label_offsets)

    # Write the labels
    for key in hash_slots:
        stored_labels = hash_slots[key]
        for label in stored_labels:
            writer.write_uint8(len(label))
            writer.write_string(label)
            writer.write_uint32(index_map[label])


#  See https://nintendo-formats.com/libs/lms/overview.html#hash-tables
def _calculate_hash(label: str, slot_count: int) -> int:
    hash = 0
    for character in label:
        hash = hash * 0x492 + ord(character)
    return (hash & 0xFFFFFFFF) % slot_count



from lms.fileio.io import FileReader
from lms.project.definitions.tag import LMS_TagDefinition


def read_tag2(reader: FileReader) -> list[LMS_TagDefinition]:
    info_list = []

    count = reader.read_uint32()
    for offset in reader.read_offset_array(count):
        reader.seek(offset)

        param_count = reader.read_uint16()
        parameter_indexes = reader.read_uint16_array(param_count)
        name = reader.read_encoded_string()

        info_list.append(LMS_TagDefinition(name, parameter_indexes))

    return info_list



from typing import Any

from lms.common.lms_fileinfo import LMS_FileInfo
from lms.message.msbtentry import MSBTEntry
from lms.titleconfig.definitions.attribute import AttributeConfig
from lms.titleconfig.definitions.tags import TagConfig


class MSBT:
    """
    A class that represents a MSBT file.

    https://nintendo-formats.com/libs/lms/msbt.html.
    """

    MAGIC = "MsgStdBn"

    # While 101 is the default slot count for LBL1 sections in a MSBT
    # The value may be overridden at the instance level as some games alter the value
    DEFAULT_SLOT_COUNT = 101

    def __init__(
        self,
        info: LMS_FileInfo | None = None,
        section_list: list[str] | None = None,
        _unsupported_section_map: dict[str, bytes] | None = None,
        attribute_config: AttributeConfig | None = None,
        tag_config: TagConfig | None = None,
    ):
        self._info = info if info is not None else LMS_FileInfo()

        self._entries: list[MSBTEntry] = []
        self._label_map: dict[str, MSBTEntry] = {}

        self.size_per_attribute = 0

        self.slot_count = MSBT.DEFAULT_SLOT_COUNT

        self.uses_encoded_attributes = True
        self.attr_string_table: bytes | None = None

        self._unsupported_section_map = _unsupported_section_map or {}

        # Store the section list so that the order of any and all sections is preserved when writing
        self._section_list: list[str] = section_list or ["LBL1"]

        self._attribute_config = attribute_config
        self._tag_config = tag_config

    def __iter__(self):
        return iter(self._entries)

    @property
    def info(self) -> LMS_FileInfo:
        """The file info for the MSBT instance."""
        return self._info

    @property
    def section_list(self) -> tuple[str, ...]:
        """The list of sections with order preserved."""
        return tuple(self._section_list)

    @property
    def unsupported_sections(self) -> tuple[str, ...]:
        """The list of unsupported sections."""
        return tuple(self._unsupported_section_map.keys())

    @property
    def has_attributes(self) -> bool:
        """If the msbt contains attributes."""
        return self.section_exists("ATR1")

    @property
    def has_style_indexes(self) -> bool:
        """If the msbt contains style indexes."""
        return self.section_exists("TSY1")

    def get_entry_by_index(self, index: int) -> MSBTEntry:
        """
        Retrieves an entry given its index. Supports negative indexing.

        :param index: the index of the entry.
        """
        if not (-len(self._entries) <= index < len(self._entries)):
            raise IndexError(f"The index {index} is not a valid MSBT entry!")

        return self._entries[index]

    def get_entry_by_name(self, label: str) -> MSBTEntry:
        """
        Retrieves an entry given its name.

        :param label: the label name for the entry.
        """
        if label not in self._label_map:
            raise KeyError(f"The label '{label}' does not exist!")

        return self._label_map[label]

    def add_entry(self, entry: MSBTEntry) -> None:
        """
        Adds an entry to the MSBT instance.

        :param entry: the MSBTEntry object to add.
        """
        if entry.name in self._label_map:
            raise KeyError(f"The label '{entry.name}' already exists!")

        if self.section_exists("ATR1"):
            if entry.attribute is None:
                raise ValueError(
                    f"Entry '{entry.name}' can't be added with no attributes when attributes already exist!"
                )
        elif entry.attribute is not None:
            self._section_list.insert(1, "ATR1")
            self._info.section_count += 1

        # TXT2 will always exist so insert it at all times
        if not self.section_exists("TXT2"):
            self._section_list.insert(2, "TXT2")

        if self.section_exists("TSY1"):
            if entry.style_index is None:
                raise ValueError(
                    f"Entry '{entry.name}' can't be added with no style index when styles already exist!"
                )
        elif entry.style_index is not None:
            self._section_list.insert(3, "TSY1")
            self._info.section_count += 1

        self._entries.append(entry)
        self._label_map[entry.name] = entry

    def delete_entry(self, entry: MSBTEntry) -> None:
        """
        Removes an entry from the MSBT instance.

        :param entry: the MSBTEntry object to remove.
        """
        if entry.name not in self._label_map:
            raise KeyError(f"The entry '{entry.name}' does not exist!")

        self._entries.remove(entry)
        del self._label_map[entry.name]

    def section_exists(self, name: str) -> bool:
        """
        Determines if a section exists in the current MSBT.

        :param name: the name of the section.
        """
        return name in self._section_list

    def get_unsupported_section_data(self, name: str) -> bytes:
        """
        Retrieves the raw data of an unsupported section.

        :param name: the name of the section.
        """
        if name not in self._unsupported_section_map:
            raise KeyError(f"The section '{name}' does not exist in the MSBT!")

        return self._unsupported_section_map[name]
